package csci1902.hw3;

import junit.framework.TestCase;
import java.awt.Point;
import java.util.*;

/**
 * Test cases for common ant functionality
 * @author Michael Ekstrand <ekstrand@cs.umn.edu>
 *
 */
public class ForagerAntTest extends TestCase {
    private World world;
    private ForagerAnt ant;
    private FakeRandomGenerator rng;
    
    /**
     * Construct a new forager ant to test its behavior.
     */
    protected void setUp() {
        rng = new FakeRandomGenerator();
        world = new World(10, 10);
        ant = new ForagerAnt(20, world, new Point(5,7));
        ant.setPosition(new Point(1,1));
        world.setRandomGenerator(rng);
    }
    
    /**
     * Make sure that the ant's initial state is correct (position, home
     * position, food).
     */
    public void testInitialState() {
        assertSame(world, ant.getWorld());
        assertEquals(0, ant.getFoodUnits());
        assertEquals(new Point(5,7), ant.getHomePosition());
    }
    
    /**
     * Make sure that the ant moves randomly correctly.  This test method
     * depends on knowledge of how the moveRandomly() method works, and preseeds
     * the fake RNG to get a predictable movement.
     */
    public void testRandomMovement() {
        rng.addNumber(0.05); // northwest motion
        ant.act(new ArrayList<Entity>());
        assertEquals(new Point(1,1), ant.getPosition());
        world.commit();
        assertEquals(0, world.getFoodPheromone(1,1), 0.0001);
        
        ant.update();
        assertEquals(0, world.getFoodPheromone(1,1), 0.0001);
        assertEquals(new Point(0,0), ant.getPosition());
        assertEquals(0, world.getFoodPheromone(0,0), 0.0001);
        rng.addNumber(.05);  // east motion in this case - will be the first
        ant.act(new ArrayList<Entity>());
        ant.update();
        assertEquals(new Point(1, 0), ant.getPosition());
        assertEquals(0, world.getFoodPheromone(1,0), 0.0001);
    }
    
    /**
     * Test the pheremone-tracking code.
     */
    public void testPheremoneTrack() {
        world.depositFoodPheromone(2, 1, 1);
        world.depositFoodPheromone(2, 0, 2);
        world.depositFoodPheromone(0, 0, 2);
        world.depositFoodPheromone(1, 2, 3);
        world.commit();
        rng.setDefault(0.7);
        ant.act(new ArrayList<Entity>());
        ant.update();
        assertEquals(new Point(2, 1), ant.getPosition());
    }
    
    /**
     * Test that an ant collects food.
     */
    public void testCollectFood() {
        FoodSource fs = new FoodSource(world, new Point(2,2), 100);
        List<Entity> neighbors = new ArrayList<Entity>();
        neighbors.add(fs);
        ant.act(neighbors);
        assertEquals(0, ant.getFoodUnits());
        ant.update();
        fs.update();
        assertEquals(20, ant.getFoodUnits());
        assertEquals(80, fs.getFoodUnits());
        assertEquals(new Point(1,1), ant.getPosition());
    }
    
    /**
     * Gives the ant some food for interactions with food sinks, etc.
     */
    private void feedAnt() {
        FoodSource fs = new FoodSource(world, new Point(2,2), 100);
        List<Entity> neighbors = new ArrayList<Entity>();
        neighbors.add(fs);
        ant.act(neighbors);
        ant.update();
        fs.update();
    }
    
    /**
     * Test that an ant deposits food.
     */
    public void testDepositFood() {
        feedAnt();
        
        Colony c = new Colony(world, new Point(0, 1));
        List<Entity> neighbors = new ArrayList<Entity>();
        neighbors.add(c);
        ant.act(neighbors);
        assertEquals(20, ant.getFoodUnits());
        ant.update();
        c.update();
        assertEquals(0, ant.getFoodUnits());
        assertEquals(20, c.getFoodLevel());
        assertEquals(new Point(1,1), ant.getPosition());
    }
    
    /**
     * Test that an ant deposits pheromone as it moves while carrying food.
     */
    public void testDepositPheromone() {
        feedAnt();
        
        ant.act(new ArrayList<Entity>());
        assertEquals(new Point(1,1), ant.getPosition());
        ant.update();
        world.commit();
        // make sure that the ant deposits pheromone where it was
        assertEquals(5, world.getFoodPheromone(1, 1), 0.0001);
        assertEquals(0, world.getFoodPheromone(2, 2), 0.0001);
    }
    
    /**
     * Test that an ant goes home when carrying food.
     */
    public void testGoHome() {
        feedAnt();
        List<Entity> lst = new ArrayList<Entity>();
        ant.act(lst);
        assertEquals(new Point(1,1), ant.getPosition());
        ant.update();
        assertEquals(new Point(2, 2), ant.getPosition());
        ant.act(lst);
        ant.update();
        assertEquals(new Point(3,3), ant.getPosition());
        ant.act(lst);
        ant.update();
        assertEquals(new Point(3,4), ant.getPosition());
        ant.act(lst);
        ant.update();
        assertEquals(new Point(4,5), ant.getPosition());
        ant.act(lst);
        ant.update();
        assertEquals(new Point(4,6), ant.getPosition());
        ant.act(lst);
        ant.update();
        assertEquals(new Point(5,7), ant.getPosition());
    }
}
