package csci1902.hw3;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

import javax.swing.*;
import javax.swing.Timer;
import javax.swing.event.*;
import javax.imageio.*;

/**
 * Main class implementing the simulator UI and run.
 * @author Michael Ekstrand <ekstrand@cs.umn.edu>
 *
 */
@SuppressWarnings("serial")
public class Simulator extends JFrame {
    
    private World theWorld;
    
    private JToolBar toolbar;
    private JButton bStep;
    private JToggleButton bRun;
    private JSlider simDelay;
    private Timer runTimer;
    private SimComponent view;
    
    /**
     * Create a new simulator window.
     */
    public Simulator() {
        this(null);
    }
    
    /**
     * Create a new simulator window with a file loaded.
     * @param filename The file to load a world from.
     */
    public Simulator(String filename) {
        setLayout(new BorderLayout());
        setDefaultCloseOperation(EXIT_ON_CLOSE);
        setTitle("Ant Simulation");
        
        runTimer = new Timer(100, new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                onStep();
            }
        });
        
        toolbar = new JToolBar();
        add(toolbar, BorderLayout.NORTH);
        
        JButton open = new JButton("Open");
        open.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                onOpen();
            }
        });
        open.setToolTipText("Load a simulation file");
        toolbar.add(open);
        toolbar.addSeparator();
        bStep = new JButton("Step");
        bStep.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                onStep();
            }
        });
        bStep.setToolTipText("Run the simulation one step");
        toolbar.add(bStep);
        bRun = new JToggleButton("Run");
        bRun.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent e) {
                onRunToggled();
            }
        });
        bRun.setToolTipText("Run the simulation continuously");
        toolbar.add(bRun);
        simDelay = new JSlider(1, 20, 10);
        simDelay.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent e) {
                onRunDelayChanged();
            }
        });
        simDelay.setSnapToTicks(true);
        simDelay.setMinorTickSpacing(1);
        simDelay.setMajorTickSpacing(10);
        simDelay.setToolTipText("Adjust the simulation speed");
        toolbar.add(simDelay);
        toolbar.addSeparator();
        JButton quit = new JButton("Quit");
        quit.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                System.exit(0);
            }
        });
        quit.setToolTipText("Quit the simulator");
        toolbar.add(quit);
        
        view = new SimComponent(null);
        add(view, BorderLayout.CENTER);
        
        if (filename == null)
            setWorld(null);
        else
            load(null, new File(filename));
        
        pack();
    }
    
    private void setWorld(World world) {
        runTimer.stop();
        theWorld = world;
        bStep.setEnabled(world != null);
        bRun.setEnabled(world != null);
        view.setWorld(world);
        pack();
    }
    
    /**
     * Handle clicks on the Open button.
     */
    private void onOpen() {
        JFileChooser chooser = new JFileChooser();
        int rv = chooser.showOpenDialog(this);
        if (rv == JFileChooser.APPROVE_OPTION) {
            load(this, chooser.getSelectedFile());
        }
    }
    
    /**
     * Run the simulation one step.
     */
    private void onStep() {
        theWorld.runSimulationStep();
        view.repaint();
    }
    
    /**
     * Toggle autorunning of the simulation.
     */
    private void onRunToggled() {
        if (theWorld != null && bRun.isSelected()) {
            runTimer.start();
            bStep.setEnabled(false);
        } else {
            runTimer.stop();
            bStep.setEnabled(true);
        }
    }
    
    /**
     * Handle change of the simulation speed timer.
     */
    private void onRunDelayChanged() {
        runTimer.setDelay(1000 / simDelay.getValue());
    }
    
    /**
     * Load a world file.
     * @param errParent The parent component for error messages.
     * @param file The file to load.
     */
    private void load(Component errParent, File file) {
        try {
            World w = readWorld(file);
            setWorld(w);
        } catch (IOException e) {
            JOptionPane.showMessageDialog(errParent,
                    String.format("Error occured opening %s\n%s",
                            file.getName(), e.getMessage()),
                    "Error",
                    JOptionPane.ERROR_MESSAGE);
        }
    }
    
    /**
     * Reads in a new world file.
     * @param file The file to read from
     * @return The world object
     * @throws IOException If there is an error reading the file.
     */
    private World readWorld(File file) throws IOException {
        Scanner s = new Scanner(file);
        
        try {
            if (!s.next().equals("world")) {
                throw new InvalidWorldException("File must begin with 'world'");
            }
            int width = s.nextInt();
            int height = s.nextInt();
            if (!s.next().equals("colony")) {
                throw new InvalidWorldException(
                        "Colony must immediately follow world");
            }
            Point cpos = new Point();
            cpos.x = s.nextInt();
            cpos.y = s.nextInt();
            
            World world = new World(width, height, cpos);
            
            while (s.hasNext()) {
                String cmd = s.next();
                if (cmd.equals("food")) {
                    int quantity = s.nextInt();
                    Point pos = new Point();
                    pos.x = s.nextInt();
                    pos.y = s.nextInt();
                    world.addEntity(new FoodSource(world, pos, quantity));
                } else if (cmd.equals("waste")) {
                    Point pos = new Point();
                    pos.x = s.nextInt();
                    pos.y = s.nextInt();
                    world.addEntity(new WasteDump(world, pos));
                } else if (cmd.equals("forager-ants")) {
                    int n = s.nextInt();
                    for (int i = 0; i < n; i++) {
                        world.addForagerAnt();
                    }
                } else if (cmd.equals("waste-ants")) {
                    int n = s.nextInt();
                    for (int i = 0; i < n; i++) {
                        world.addWasteAnt();
                    }
                } else if (cmd.equals("spider")) {
                    Point pos = new Point();
                    pos.x = s.nextInt();
                    pos.y = s.nextInt();
                    // TODO Uncomment this line to add the spider
                    // world.addEntity(new Spider(world, pos));
                } else {
                    throw new InvalidWorldException("Invalid command: " + cmd);
                }
            }

            return world;
        } catch (NoSuchElementException e) {
            throw new EOFException();
        } finally {
            s.close();
        }
    }

    /**
     * Main entry point for the simulator.
     * @param args
     */
    public static void main(String[] args) {
        Simulator sim;
        if (args.length > 0) {
            sim = new Simulator(args[0]);
        } else {
            sim = new Simulator();
        }
        sim.setVisible(true);
    }
    
    /**
     * A component for displaying the simulator.
     * @author Michael Ekstrand <ekstrand@cs.umn.edu>
     *
     */
    private static class SimComponent extends JComponent {
        private World theWorld;
        
        private Map<Class, BufferedImage> images =
            new HashMap<Class,BufferedImage>();
        
        public SimComponent(World world) {
            loadImages();
            setWorld(world);
        }
        
        /**
         * Set the world to be displayed by this view.
         * @param world The world to display.
         */
        public void setWorld(World world) {
            theWorld = world;
            if (world != null) {
                Dimension d = new Dimension(
                        world.getWidth() * 8, 
                        world.getHeight() * 8);
                setMinimumSize(d);
                setPreferredSize(d);
                invalidate();
            } else {
                Dimension d = new Dimension(100, 100);
                setMinimumSize(d);
                setPreferredSize(d);
            }
        }
        
        protected void paintComponent(Graphics g) {
            if (theWorld == null)
                return;
            
            Graphics2D g2 = (Graphics2D) g;
            
            for (int x = 0; x < theWorld.getWidth(); x++) {
                for (int y = 0; y < theWorld.getHeight(); y++) {
                    Graphics2D pg = (Graphics2D) g2.create(x*8, y*8, 8, 8);
                    Color c = makePheromoneColor(theWorld.getFoodPheromone(x, y));
                    pg.setColor(c);
                    Polygon poly = new Polygon();
                    poly.addPoint(0,0);
                    poly.addPoint(0,8);
                    poly.addPoint(8, 0);
                    pg.fill(poly);
                    pg.setColor(
                            makePheromoneColor(theWorld.getWastePheromone(x,y)));
                    poly.reset();
                    poly.addPoint(0,8);
                    poly.addPoint(8,0);
                    poly.addPoint(8,8);
                    pg.fill(poly);
                }
            }
            
            for (Entity e: theWorld.getEntities()) {
                Point pos = e.getPosition();
                BufferedImage img = images.get(e.getClass());
                if (img == null) {
                    System.out.println("Warning: couldn't look up image");
                } else {
                    g2.drawImage(img, null, pos.x*8, pos.y*8);
                }
            }
        }
        
        private Color makePheromoneColor(double level) {
            if (level > 5)
                level = 5;
            level *= 0.2;
            return new Color(0.0f, 0.0f, (float) level, 1.0f);
        }
        
        /**
         * Load the images for drawing the display.
         */
        private void loadImages() {
            BufferedImage src = null;
            try {
                src = ImageIO.read(new File("imgs.png"));
            } catch (IOException e) {
                System.err.println("Cannot find image resource file");
                System.exit(1);
            }
            
            loadImage(src, "csci1902.hw3.ForagerAnt", 0, 0);
            loadImage(src, "csci1902.hw3.FoodSource", 9, 0);
            loadImage(src, "csci1902.hw3.Colony", 18, 0);
            loadImage(src, "csci1902.hw3.WasteTransportAnt", 0, 9);
            loadImage(src, "csci1902.hw3.WasteDump", 9, 9);
            loadImage(src, "csci1902.hw3.Spider", 18, 9);
        }
        
        /**
         * Load a particular image
         * @param src The base source image
         * @param name The name of the class with which to associate the image
         * @param x The X coordinate of the start of the imge
         * @param y The Y coordinate of the start of the image
         */
        private void loadImage(BufferedImage src, String name, int x, int y) {
            BufferedImage subimg = src.getSubimage(x, y, 8, 8);
            try {
                Class c = Class.forName(name);
                images.put(c, subimg);
            } catch (ClassNotFoundException e) {
                System.err.format("Warning: couldn't find '%s'\n", name);
            }
        }
    }
}
