package csci1902.hw3;

import java.util.*;
import java.awt.Point;

/**
 * Class implementing the world (particularly the ground) of the simulation.
 * @author Michael Ekstrand <ekstrand@cs.umn.edu>
 *
 */
public class World {
    private double[][] foodLevels, nextFoodLevels;
    
    private int width, height;
    private RandomGenerator random = new RealRandomGenerator();
    private int foodHalfLife = 5;
    private List<Entity> entities = new ArrayList<Entity>();
    private Map<Point, List<Entity>> cells = null;
    private Colony colony;
    
    /**
     * Builds a world with the specified width and height.
     * @param w The width of the world
     * @param h The height of the world
     * @param colony The colony location.  If null, puts the colony at 0,0.
     */
    public World(int w, int h, Point cpos) {
        width = w;
        height = h;
        foodLevels = new double[w][h];
        nextFoodLevels = new double[w][h];
        for (int i = 0; i < w; i++) {
            for (int j = 0; j < h; j++) {
                foodLevels[i][j] = 0;
                nextFoodLevels[i][j] = 0;
            }
        }
        
        if (cpos == null)
            cpos = new Point(0,0);
        colony = new Colony(this, cpos);
        addEntity(colony);
    }
    
    /**
     * Creates a world with a colony at 0,0
     * @param w The world's width
     * @param h The world's height
     */
    public World(int w, int h) {
        this(w, h, null);
    }
    
    /**
     * Run one step of the simulation.
     */
    public void runSimulationStep() {
        // Invoke entity state preparation
        for (Entity e: entities) {
            List<Entity> neighbors = collectNeighbors(e);
            e.act(neighbors);
        }
        // Commit states
        for (Entity e: entities) {
            e.update();
        }
        commit();
        // Clear out our cache of entity cells, as they're likely invalid
        cells = null;
    }
    
    /**
     * Get the active random number generator.
     * @return the random number generator
     */
    public RandomGenerator getRandomGenerator() {
        return random;
    }
    /**
     * Override the random generator.  Only used for testing.
     * @param g The random number generator to use.
     */
    public void setRandomGenerator(RandomGenerator g) {
        random = g;
    }
    
    /**
     * Commit changes to create the next state of the world.  This method is
     * called after all entities have been committed.
     */
    public void commit() {
        // Copy the next food levels into the current state
        for (int i = 0; i < width; i++) {
            for (int j = 0; j < height; j++) {
                foodLevels[i][j] = nextFoodLevels[i][j];
            }
        }
        // Apply decay for the next state (so it's ready to go)
        applyDecay();
        // Delete dead entities
        Iterator<Entity> eit = entities.iterator();
        while (eit.hasNext()) {
            Entity e = eit.next();
            if (!e.isAlive()) {
                eit.remove();
            }
        }
    }
    
    /**
     * Queries the width of the world.
     * @return The width of the world.
     */
    public int getWidth() {
        return width;
    }
    
    /**
     * Queries the height of the world.
     * @return The height of the world.
     */
    public int getHeight() {    
        return height;
    }
    
    /**
     * Retrieve the world's colony.
     * @return The colony
     */
    public Colony getColony() {
        return colony;
    }
    
    /**
     * Return the array of active entities in the world.
     * @return the array of active entities in the world.
     */
    public List<Entity> getEntities() {
        // return a copy so our caller can't modify our list without our
        // knowledge
        return new ArrayList<Entity>(entities);
    }
    
    /**
     * Add a forager ant to the world.  The ant is positioned at the colony.
     */
    public void addForagerAnt() {
        ForagerAnt ant = new ForagerAnt(20, this, colony.getPosition());
        addEntity(ant);
    }
    
    /**
     * Add a waste ant to the world.  The ant is positioned at the colony.
     */
    public void addWasteAnt() {
        // TODO: Add code here to add a waste ant with capacity 20 in the B level
    }
    
    public void addEntity(Entity obj) {
        entities.add(obj);
    }
    
    /**
     * Retrieve a list of cells neighboring a particular cell.
     * @param pos The cell whose neighbors are desired
     * @return A list of points representing cells neighboring {@code pos}
     */
    public List<Point> getNeighboringCells(Point pos) {
        ArrayList<Point> positions = new ArrayList<Point>(8);
        for (int i = -1; i <= 1; i++) {
            for (int j = -1; j <= 1; j++) {
                if (i != 0 || j != 0) {
                    if (pos.x + j >= 0 && pos.y + i >= 0
                            && pos.x + j < width && pos.y + i < height) {
                        positions.add(new Point(pos.x + j, pos.y + i));
                    }
                }
            }
        }
        return positions;
    }
    
    /**
     * Deposit an amount of food pheremone in the specified cell.
     * @param x The cell's X coordinate.
     * @param y The cell's Y coordinate.
     * @param amount The (positive) amount of food pheremone to deposit.
     */
    public void depositFoodPheromone(int x, int y, double amount) {
        nextFoodLevels[x][y] += amount;
    }
    
    /**
     * Get the food pheremone level of a particular cell.
     * @param x The cell's X coordinate.
     * @param y The cell's Y coordinate.
     * @return The amount of food pheremone present in the cell.
     */
    public double getFoodPheromone(int x, int y) {
        return foodLevels[x][y];
    }
    
    /**
     * Deposit waste pheromone at the position (x,y).
     * @param x The X-coordinate of the cell of interest
     * @param y The Y-coordinate of the cell of interest
     * @param amount The amount of pheromone to deposit.
     */
    public void depositWastePheromone(int x, int y, double amount) {
        // TODO: Implement this method for the B level
    }
    /**
     * Get the amount of waste pheromone at the position (x,y).
     * @param x The X-coordinate of the cell of interest
     * @param y The Y-coordinate of the cell of interest
     * @return The amount of waste pheromone at the cell
     */
    public double getWastePheromone(int x, int y) {
        // TODO: Make this do something real for the B level
        return 0;
    }
   
    /**
     * Applies decay to the next food levels.
     */
    private void applyDecay() {
        for (int i = 0; i < width; i++) {
            for (int j = 0; j < height; j++) {
                nextFoodLevels[i][j] *= decayMultiplier(foodHalfLife);
            }
        }
    }
    
    /**
     * Convert a half-life to a decay multiplier.
     * @param halflife The half-life (in time steps) of the decay curve.
     * @return The multipler which, when multiplied by a value, will decay
     * that value for one time step.
     */
    private double decayMultiplier(int halflife) {
        double exponent = Math.log(0.5) / halflife;
        return Math.exp(exponent);
    }
    
    /**
     * Collect the neighbors of an entity.
     * @param entity The entity whose neighbors are desired
     * @return A list containing all neighbors of the entity
     */
    private List<Entity> collectNeighbors(Entity entity) {
        Point pos = entity.getPosition();
        
        if (cells == null) {
            gatherCells();
        }
        
        // grab the neighboring cells and add the cell itself
        List<Point> searchCells = getNeighboringCells(pos);
        searchCells.add(pos);
        
        List<Entity> neighbors = new ArrayList<Entity>();
        
        // go over each of the cells in the serach space, gathering entities
        // from the cache
        for (Point p: searchCells) {
            List<Entity> l = cells.get(p);
            if (l != null)
                neighbors.addAll(l);
        }
        
        // Remove the entity itself from its neighbor list
        neighbors.remove(entity);

        return neighbors;
    }
    
    /**
     * Group the entities by cell into the {@link #cells} cache.
     */
    private void gatherCells() {
        cells = new HashMap<Point, List<Entity>>();
        for (Entity e: entities) {
            Point key = e.getPosition();
            if (!cells.containsKey(key))
                cells.put(key, new LinkedList<Entity>());
            
            cells.get(key).add(e);
        }
    }
}
